/***********************************************************
 * $Id: POJOComparatorFactory.java 36 2007-03-29 15:52:53Z wglas $
 * 
 * Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: Feb 28, 2007
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 * 
 ***********************************************************/

package org.clazzes.util.comparators;

import java.util.Comparator;
import java.util.Locale;

import org.apache.commons.collections.comparators.ComparatorChain;
import org.apache.commons.collections.comparators.ReverseComparator;
import org.clazzes.util.reflect.FieldAccessor;
import org.clazzes.util.reflect.POJOFieldAccessor;

/**
 * This factory contructs a comparator for a POJO class by parsing an expression.
 * 
 * @author wglas
 */
public class POJOComparatorFactory {

    private ComparatorFactory comparatorFactory;
    
    /**
     * Contruct a POJOComparatorFactory using an instance of
     * {@link DefaultComparatorFactory}.
     * 
     * @param comparatorFactory The factory, which constructs comparators for
     *                    individual property values.
     */
    public POJOComparatorFactory()
    {
        this.comparatorFactory = DefaultComparatorFactory.getInstance();
    }
    
    /**
     * Contruct a POJOComparatorFactory using an instance of
     * {@link LocalizedComparatorFactory}
     * 
     * @param comparatorFactory The factory, which constructs comparators for
     *                    individual property values.
     */
    public POJOComparatorFactory(Locale locale)
    {
        this.comparatorFactory = new LocalizedComparatorFactory(locale);
    }
    
    /**
     * Contruct a POJOComparatorFactory using the given comparator factory
     * for constructing value comparators.
     * 
     * @param comparatorFactory The factory, which constructs comparators for
     *                    individual property values.
     */
    public POJOComparatorFactory(ComparatorFactory comparatorFactory)
    {
        this.comparatorFactory = comparatorFactory;
    }
    
    /**
     * find the next horzonal bar character at the outermost brace level.
     * 
     * @param s The string in which to search.
     * @param start The start position.
     * @return The position of the next horizontal character.
     */
    private static int nextHor(String s, int start)
    {
        int depth = 0;
        
        int pos = start;
        
        while (pos < s.length())
        {
            char c = s.charAt(pos);
            
            switch (c)
            {
            case '{':
                ++depth;
                break;
            case '}':
                --depth;
                if (depth < 0) return -1;
                break;
            case '|':
                if (depth==0) return pos;
                break;
            }
            ++pos;
        }
        return -1;
    }
    
    
    /**
     * Construct a comparator for the given class by evaluating the specified chain
     * of properties.
     * 
     * @param clazz The class of the POJO object, which will be compared by the returned comparator.
     * @param expression An expression, which represnets an OGNL-like chain of properties
     *                   like the following examples:
     * <ul>
     * <li>myProperty</li> Compare <code>x.getMyProperty()</code> by <code>y.getMyProperty()</code>.
     * <li>myProperty.abc</li> Compare <code>x.getMyProperty().getAbc()</code> by <code>y.getMyProperty().getAbc()</code>.
     * <li>{abc|cde|ghi}</li> Lexicographically compare the values of <code>abc</code>,
     *                       <code>def</code> and <code>ghi</code>. The first property is the most significant.
     * <li>prop.{abc|cde|ghi}</li> Lexicographically compare the values of <code>prop.abc</code>,
     *                       <code>prop.def</code> and <code>gprop.hi</code>.
     * </ul>
     * 
     * Each component may be preceded by a minus sign in order to invert the sort order:
     * 
     * <ul>
     * <li>-prop</li> Reverse compare the values of <code>prop</code>.
     * <li>-{abc|-cde|ghi}</li> Lexicographically reverse compare the values of <code>abc</code>,
     *                       <code>-def</code> and <code>ghi</code>. The first property is the most significant.
     * </ul>
     * 
     * @param ascending If <code>false</code>, his parameter my be used to invert the sort oder
     *                  of the returned comparator.
     * 
     * @return An comparator for the given POJO class using the given expression.
     * @throws InvalidExpressionSyntaxException If the syntax of the expression is invalid.
     */
    public Comparator getComparator(Class clazz, String expression, boolean ascending) throws InvalidExpressionSyntaxException
    {
        try {
            String exp = expression.trim();
            
            boolean invert = exp.charAt(0) == '-';
            
            if (invert)
            {
                exp = exp.substring(1).trim();
                ascending = !ascending;
            }
            
            // split up {exp1|exp2|...} contructs, which may only appear at the end of an expresion.
            if (exp.startsWith("{"))
            {
                if(!exp.endsWith("}"))
                    throw new InvalidExpressionSyntaxException(expression,"The expression starts with '{' but does not end with '}'");
                
                ComparatorChain chain=null;
                
                int lastPos = 1;
                int nextPos;
                
                while ((nextPos = POJOComparatorFactory.nextHor(exp,lastPos)) > lastPos)
                {
                   if (chain == null) chain = new ComparatorChain();
                
                    chain.addComparator(this.getComparator(clazz, exp.substring(lastPos,nextPos),true),!ascending);
                    
                    lastPos = nextPos +1;
                }
                
                if (chain == null)
                    return this.getComparator(clazz, exp.substring(lastPos,exp.length()-1),ascending);
                
                chain.addComparator(this.getComparator(clazz, exp.substring(lastPos,exp.length()-1),true),!ascending);
                
                return chain;
            }
            
            // split expression at dots.
            int pos = exp.indexOf('.');

            Comparator comp;
            
            if (pos >= 0)
            {
                FieldAccessor leftAccessor = new POJOFieldAccessor(clazz,exp.substring(0,pos).trim());
                
                Comparator right = this.getComparator(leftAccessor.getType(), exp.substring(pos+1),true);
            
                comp = new POJOFieldComparator(leftAccessor,right);
            }
            else
            {
                comp = new POJOFieldComparator(clazz,exp,this.comparatorFactory);
            }
            
            if (ascending)
                return comp;
            else
                return new ReverseComparator(comp);
            
        } catch (SecurityException e) {
            throw new InvalidExpressionSyntaxException(expression,"SecurityException fetching field getter.",e);
        } catch (NoSuchMethodException e) {
            throw new InvalidExpressionSyntaxException(expression,"NoSuchMethodException fetching field getter.",e);
        }
        
    }
}
